/* my fileio */

#include <basic.h>
#include <bstdio.h>
#include <stdarg.h>
#include <tstring.h>
#include <new>
#include <tcode.h>

#define FILEIO_DEFINE
#include "fileio.h"

// nothing as of yet;

static FILE stdout_instance = { DT_STDOUT, OP_READ, 0, 0 };
static FILE stderr_instance = { DT_STDERR, OP_READ, 0, 0 };

FILE *stdout = &stdout_instance;
FILE *stderr = &stderr_instance;

int fprintf(FILE *stream, const char *format, ...)
{
  int ret = -1;
  va_list args;

  switch (stream->type) {
      case DT_STDOUT:
      case DT_STDERR:
        va_start(args, format);
        ret = vprintf(format, args);
        va_end(args);
        break;
      case DT_FILE:
        /* TODO */

      default:
        printf("fprintf: Unimplemented as of yet\n");
        break;
  }

  return ret;
}

FILE *fopen(const char *path, const char *mode)
{
  FILE *fp = 0;
  TC *tcpath = 0;
  LINK link;
  int len;
  W err;

  switch (mode[0]) {
      case 'r': // r, rb, r+
        if (mode[1] == '\0' || mode[1] == 'b') {
          try {
            fp = new FILE;
            fp->type = DT_FILE;
            fp->ope = OP_READ;
            fp->fd = -1;
            fp->br = 0;
            len = strlen(path);
            tcpath = new TC[len+1];
            eucstotcs(tcpath, path);
            for (int i = 0; i < len; i++) {
              if (tcpath[i] == TK_SLSH) tcpath[i] = TC_FDLM;
            }
            if ((err = get_lnk(tcpath, &link, F_NORM)) < 0) {
              goto error1;
            }
            if ((fp->fd = opn_fil(&link, F_READ, NULL)) < 0) {
              goto error1;
            }
            // rectype 31 only
            if ((err = fnd_rec(fp->fd, F_TOPEND, 1<<31, 0, NULL)) < 0) {
              goto error1;
            }
            fp->br = new BufferedReader(fp->fd);

            delete [] tcpath;
            tcpath = 0;
            break;

          error1:
            delete fp;
            delete [] tcpath;
            fp = 0;
            tcpath = 0;
          }
          catch (std::bad_alloc e) {
            if (fp != 0) {
              delete fp->br;
              if (fp->fd >= 0) {
                cls_fil(fp->fd);
              }
            }
            delete fp;
            delete [] tcpath;
            return 0;
          }
        }
        else {
          printf("fopen: Unimplemented as of yet\n");
        }
        break;
      case 'w': // w, w+
      case 'a': // a, a+
        /* TODO */

      default:
        printf("fopen: Unimplemented as of yet\n");
        break;
  }

  return fp;
}

FILE *fopen_link(const LINK* link, const char *mode)
{
  FILE *fp = 0;
  W err;

  switch (mode[0]) {
      case 'r': // r, rb, r+
        if (mode[1] == '\0' || mode[1] == 'b') {
          try {
            fp = new FILE;
            fp->type = DT_FILE;
            fp->ope = OP_READ;
            fp->fd = -1;
            fp->br = 0;
            if ((fp->fd = opn_fil(const_cast<LINK*>(link), 
                                  F_READ, NULL)) < 0) {
              goto error1;
            }
            // rectype 31 only
            if ((err = fnd_rec(fp->fd, F_TOPEND, 1<<31, 0, NULL)) < 0) {
              goto error1;
            }
            fp->br = new BufferedReader(fp->fd);
            break;

          error1:
            delete fp;
            fp = 0;
          }
          catch (std::bad_alloc e) {
            if (fp != 0) {
              delete fp->br;
              if (fp->fd >= 0) {
                cls_fil(fp->fd);
              }
            }
            delete fp;
            return 0;
          }
        }
        else {
          printf("fopen_link: Unimplemented as of yet\n");
        }
        break;
      case 'w': // w, w+
      case 'a': // a, a+
        /* TODO */

      default:
        printf("fopen_link: Unimplemented as of yet\n");
        break;
  }

  return fp;
}

int fclose(FILE *stream)
{
  int ret = -1; // EOF

  switch (stream->type) {
      case DT_STDOUT:
      case DT_STDERR:
        ret = 0;
        break;
      case DT_FILE:
        switch (stream->ope) {
            case OP_READ:
              delete (stream->br);
              cls_fil(stream->fd);
              delete stream;
              ret = 0;
              break;
            case OP_WRITE:
              /* TODO */

            default:
              printf("fclose: Unimplemented as of yet\n");
              break;
        }
        break;
      default:
        printf("fclose: Unimplemented as of yet\n");
        break;
  }

  return ret;
}

int fflush(FILE *stream)
{
  return 0;
}

int fseeko(FILE *stream, off_t offset, int whence)
{
  return fseek(stream, offset, whence);
}

off_t ftello(FILE *stream)
{
  return ftell(stream);
}

long ftell(FILE *stream)
{
  if (stream->type != DT_FILE) {
    return -1;
  }
  if (stream->ope != OP_READ) {
    printf("ftell: Unimplemented as of yet\n");
    return -1;
  }
  return stream->br->getPos();
}

int fseek(FILE *stream, long offset, int whence)
{
  int real_offset = 0;

  if (stream->type != DT_FILE) {
    return -1;
  }
  if (stream->ope != OP_READ) {
    printf("fseek: Unimplemented as of yet\n");
    return -1;
  }
  
  switch (whence) {
      case SEEK_SET:
        real_offset = offset - stream->br->getPos();
        break;
      case SEEK_CUR:
        real_offset = offset;
        break;
      case SEEK_END:
        real_offset = offset + stream->br->getFileSize() 
          - stream->br->getPos(); // ǤʤϤޤä...
        break;
      default:
        return -1;
  }

//  printf("fseek%d (%d) %d -> ", whence, offset, stream->br->getPos());
  if (real_offset >= 0) {
    stream->br->skip(real_offset);
  }
  else {
    stream->br->offset = stream->br->getPos() + real_offset;
#if 0  
    // ޤϤʤƤ⤿֤פ()
    if (stream->br->offset > stream->br->getFileSize()) {
      stream->br->offset = stream->br->getFileSize();
    }
#endif
    stream->br->bcur = stream->br->blen = 0;
    stream->br->renewEOFState();
  }
//  printf("%d\n", stream->br->getPos());

  return 0;
}

int fgetc(FILE *stream)
{
  unsigned char c;
  size_t sz = fread(&c, 1, 1, stream);
  if (sz <= 0) {
    return -1;
  }
  return c;
}

int fputc(int c, FILE *stream) 
{
  printf("fputc: Unimplemented as of yet\n");
  return -1;
}

size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  if (stream->type != DT_FILE) {
    return -1;
  }
  if (stream->ope != OP_READ) {
    printf("fread: Unimplemented as of yet\n");
    return -1;
  }

  return (stream->br->read(reinterpret_cast<B*>(ptr), size * nmemb)) / size;
}

int ungetc(int c, FILE *stream)
{
  printf("ungetc: Unimplemented as of yet\n");  
  return -1;
}

size_t fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  printf("fwrite: Unimplemented as of yet\n");  
  return -1;
}
