/*!
 * @file MainMenu.cc
 * @brief ᥤ˥塼()
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2004-10-31, first edition (for BSlide)
 * @date 2005-02-27, modified for Scribble
 * @date 2006-05-08, modified for MicroPDF, removed a bug
 * @date 2006-10-21, modified for MicroPDF, added vobj-menu support
 *
 * Copyright (C) 2004-2006, Yashiro Takeshi
 */

#include <basic.h>
#include <btron/hmi.h>
#include <btron/vobj.h>
#include <btron/libapp.h>
#include "MainMenu.h"
#include "tcsliteral.h"
#include "Exception.h"
#include "databox.h"
#include "VobjSel.h"

static TC m_finish[]  = {TCS_MENU_FINISH};
static TC m_save[]    = {TCS_MENU_SAVE};
static TC m_disp[]    = {TCS_MENU_DISP};
static TC m_edit[]    = {TCS_MENU_EDIT};
static TC m_page[]    = {TCS_MENU_PAGE};
static TC m_zoom[]    = {TCS_MENU_MZOOM};

static MENUITEM mitems[] = {
  {0L, 0L, 0, 0, m_finish},
  {0L, 0L, 0, 0, m_save},
  {0L, 0L, 0, 0, m_disp},
  {0L, 0L, 0, 0, m_edit},
  {0L, 0L, 0, 0, m_page},
  {0L, 0L, 0, 0, m_zoom},
  {0L, 0L, 0, 0, NULL}, // (-6)
  {0L, 0L, 0, 0, NULL}, // ¿(-5)
  {0L, 0L, 0, 0, NULL}, // ǥ(-4)
  {0L, 0L, 0, 0, NULL}, // ɥ(-3)
  {0L, 0L, 0, 0, NULL}, // ʪ(-2)
  {0L, 0L, 0, 0, NULL}, // ¹(-1)
};

MainMenu::MainMenu()
: mid(-1)
{
  mcnv.insert(std::make_pair(0x001, MN_FINISH));
  mcnv.insert(std::make_pair(0x101, MN_NEWSAVE));
  mcnv.insert(std::make_pair(0x102, MN_SAVE));
  mcnv.insert(std::make_pair(0x201, MN_REDISP));
  
  mid = mcre_men(sizeof(mitems)/sizeof(MENUITEM), mitems, NULL);
  if (mid < 0){
    throw UserException(-2);
  }
  prepareMenuItems(0);
}

void MainMenu::prepareMenuItems(VID vid)
{
  int wmidx = sizeof(mitems)/sizeof(MENUITEM);
  wget_dmn(&(mitems[wmidx-3].ptr));
  mset_itm(mid, wmidx-3, &(mitems[wmidx-3]));
  oget_vmn(vid, &(mitems[wmidx-1].ptr), &(mitems[wmidx-2].ptr), 
           &(mitems[wmidx-6].ptr), &(mitems[wmidx-5].ptr), 
           &(mitems[wmidx-4].ptr), NULL);
  mset_itm(mid, wmidx-6, &(mitems[wmidx-6]));
  mset_itm(mid, wmidx-5, &(mitems[wmidx-5]));
  mset_itm(mid, wmidx-4, &(mitems[wmidx-4]));
  mset_itm(mid, wmidx-3, &(mitems[wmidx-3]));
  mset_itm(mid, wmidx-2, &(mitems[wmidx-2]));
  mset_itm(mid, wmidx-1, &(mitems[wmidx-1]));
}

// ˥塼ܤμμ
MainMenuItem MainMenu::getInternalID(W rid) const
{
  int wmidx = sizeof(mitems)/sizeof(MENUITEM);
  int major = rid >> 8;
  if (major == wmidx - 6){
    return MN_VOBJ;
  }
  else if (major == wmidx - 5) {
    return MN_ROBJ;
  }
  else if (major == wmidx - 4) {
    return MN_DISK;
  }
  else if (major == wmidx - 3) {
    return MN_WINDOW;
  }
  else if (major == wmidx - 2) {
    return MN_GADGET;
  }
  else if (major == wmidx - 1) {
    return MN_EXEC;
  }
  MConvMap::const_iterator p = mcnv.find(rid);
  if (p != mcnv.end()){
    return p->second;
  }
  return MN_VOID;
}

// ˥塼ɥɽ
W MainMenu::popupMenu(WEVENT* ev, const VobjSelection *vsel)
{
  prepareMenuItems(vsel->getMenuMode());

  setpointer(PS_SELECT, NULL);
  return msel_men(mid, ev->s.pos);
}

