// -------------------------------------------------
//  ȴ
//  (C) Copyright 2006/06, Yashiro Takeshi
// -------------------------------------------------

#include <basic.h>
#include <bstdio.h>
#include <btron/file.h>
#include "Vobj.h"
#include "BufferedWriter.h"
#include "BufferedReader.h"

#define VOBJSEGRECTYPE 29
#define VOBJSEGSUBTYPE 0xc957  // --- ¾ΥץꥱȤξͤ򤱤뤿 magic number

// ----- Vobj

Vobj::Vobj(VID avid, const VLINK& avlink, const VOBJSEG *avseg, 
           const PNT& pos)
  : vid(avid), vlink(avlink) 
{
  vseg = reinterpret_cast<VOBJSEG*>(new UB[SIZE_VOBJSEG(avseg->dlen)]);
  memcpy(vseg, avseg, SIZE_VOBJSEG(avseg->dlen));

  vseg->view.c.right = pos.x + (vseg->view.c.right - vseg->view.c.left);
  vseg->view.c.bottom = pos.y + (vseg->view.c.bottom - vseg->view.c.top);
  vseg->view.p.lefttop = pos;
}

Vobj::Vobj(VID avid, const VLINK& avlink, const VOBJSEG *avseg)
  : vid(avid), vlink(avlink) 
{
  vseg = reinterpret_cast<VOBJSEG*>(new UB[SIZE_VOBJSEG(avseg->dlen)]);
  memcpy(vseg, avseg, SIZE_VOBJSEG(avseg->dlen));
}

Vobj::Vobj(VID avid)
  : vid(avid), vseg(0) 
{
  UW size;

  oget_vob(vid, &vlink, NULL, 0, &size);
  vseg = reinterpret_cast<VOBJSEG*>(new UB[SIZE_VOBJSEG(size)]);
  oget_vob(vid, NULL, vseg, size, 0);
  refreshData();
}

void Vobj::move(WID wid, W vcode, RECT* org)
{
  switch (vcode) {
      case V_LTHD:
      case V_LBHD:
      case V_RTHD:
      case V_RBHD:
        orsz_vob(vid, org, V_SIZE);
        break;
      default:
        omov_vob(vid, wid, org, V_NODISP);
        break;
  }
  vseg->view = *org;
}

void Vobj::refreshData()
{
  UW size;
  W orgsize = SIZE_VOBJSEG(vseg->dlen);

  oget_vob(vid, &vlink, NULL, 0, &size);
  if (size > orgsize) {
    delete [] reinterpret_cast<UB*>(vseg);
    vseg = reinterpret_cast<VOBJSEG*>(new UB[SIZE_VOBJSEG(size)]);
  }
  oget_vob(vid, NULL, vseg, size, 0);
}

Vobj* Vobj::duplicate() const
{
  VID nvid = odup_vob(vid);
  if (nvid < 0) {
    return 0;
  }

  Vobj* vobj = new Vobj(nvid);
  vobj->refreshData();

  return vobj;
}

Vobj* Vobj::duplicateRobj() const
{
  VID nvid = onew_obj(vid, NULL);
  if (nvid < 0) {
    return 0;
  }
  RECT r = vseg->view;
  omov_vob(nvid, 0, &r, V_NODISP);

  Vobj* vobj = new Vobj(nvid);
  vobj->refreshData();

  return vobj;
}

Vobj* Vobj::duplicateRobj(WID wid) const
{
  // wid Υǥ̤⤷ʤΤǡcvid ͳΥԡɬ
  VID cvid = odup_vob(vid);
  omov_vob(cvid, wid, NULL, V_NODISP);
  VID nvid = onew_obj(cvid, NULL);
  if (nvid < 0) {
    return 0;
  }

  RECT r = vseg->view;
  omov_vob(nvid, 0, &r, V_NODISP);

  Vobj* vobj = new Vobj(nvid);
  vobj->refreshData();

  return vobj;
}

// ----- VobjSet

VobjSet::~VobjSet()
{
  clear();
}

void VobjSet::clear()
{
  VobjPageList::const_iterator iend = vls.end();
  for (VobjPageList::iterator i = vls.begin();
       i != iend; ++i) {
    VobjList::const_iterator jend = i->end();
    for (VobjList::iterator j = i->begin();
         j != jend; ++j) {
      delete (*j);
    }
  }
  vls.clear();
}

void VobjSet::displayPage(int pg, double zoom, UW disp) const
{
  PNT scale;
  scale.x = scale.y = 256 * zoom / 100.0;

  VobjList& vl = const_cast<VobjList&>(vls.at(pg-1));

  VobjList::const_iterator iend = vl.end();

  for (VobjList::iterator i = vl.begin();
       i != iend; ++i) {
    if (zoom == 100.0) {
      odsp_vob((*i)->getVID(), NULL, disp);
    }
    else {
      odsp_vor((*i)->getVID(), NULL, disp, scale);
    }
  }
}

Vobj* VobjSet::add(WID wid, const VLINK* vlink, const VOBJSEG* vseg, 
                   const PNT& pos, int destpage, int curpage, double zoom)
{
  RECT r;
  PNT scale;

  VID vid = oreg_vob(const_cast<VLINK*>(vlink),
                     const_cast<VOBJSEG*>(vseg), wid, V_NODISP);

  r.p.lefttop = pos;
  r.c.right = r.c.left 
    + (vseg->view.c.right - vseg->view.c.left);
  r.c.bottom = r.c.top
    + (vseg->view.c.bottom - vseg->view.c.top);

  if (destpage == curpage) {
    scale.x = scale.y = 256 * zoom / 100.0;
    omov_vob(vid, wid, &r, V_NODISP); // V_ERAORG
    if (zoom == 100.0) {
      odsp_vob(vid, NULL, V_DISPALL);
    }
    else {
      odsp_vor(vid, NULL, V_DISPALL, scale);
    }
  }

  Vobj *vobj = new Vobj(vid, *vlink, vseg, pos);
  vls.at(destpage-1).push_back(vobj);

  changed = true;

  return vobj;
}

void VobjSet::add(Vobj* vobj, int page)
{
  vls.at(page-1).push_back(vobj);
}

void VobjSet::remove(VID vid, int page)
{
  VobjList& vl = vls.at(page-1);

  VobjList::iterator iend = vl.end();
  for (VobjList::iterator i = vl.begin(); i != iend; ++i) {
    if (vid == (*i)->getVID()) {
      changed = true;
      delete (*i);
      vl.erase(i);
      break;
    }
  }
}

Vobj* VobjSet::findVobj(int pg, const PNT& p)
{
  VobjList& vl = const_cast<VobjList&>(vls.at(pg-1));

  // const_reverse_iterator ȡ!= ʬ̤顼ˤʤ
  VobjList::reverse_iterator iend = vl.rend();

  for (VobjList::reverse_iterator i = vl.rbegin();
       i != iend; ++i) {
    const RECT& r = (*i)->getRect();
    if (r.c.left <= p.x && p.x < r.c.right
        && r.c.top <= p.y && p.y < r.c.bottom) {
      return *i;
    }
  }
  return 0;
}

void VobjSet::save(const LINK* link) const
{
  if (!changed) {
    return;
  }

  W fd = opn_fil(const_cast<LINK*>(link), F_WRITE, NULL);

  // --- 󥯥쥳ɤν񤭹

  W recnum;
  see_rec(fd, 0, -1, &recnum); // ü쥳ֹ push

  for (int i = 0; i < vls.size(); i++) {
    VobjList::const_iterator jend = vls[i].end();
    for (VobjList::const_iterator j = vls[i].begin(); j != jend; ++j) {
      const Vobj *vobj = *j;
      const_cast<Vobj*>(vobj)->refreshData();
      apd_rec(fd, (B*)(vobj->getVlink()), sizeof(LINK), RT_LINK, 0, 0);
    }
  }

  // --- Ť󥯥쥳ɤκ

  see_rec(fd, recnum, 1, NULL); // (Ť)ü쥳ֹ pop

  for (;;) {
    W type = fnd_rec(fd, F_NBWD, RM_LINK, 0, NULL);
    if (type < 0) {
      break;
    }
    del_rec(fd);
  }

  // --- VOBJSEG Ǽѥ쥳ɤν񤭹

  apd_rec(fd, NULL, 0, VOBJSEGRECTYPE, VOBJSEGSUBTYPE, 0);
  see_rec(fd, -1, -1, NULL);

  {
    BufferedWriter writer(fd);

    for (int i = 0; i < vls.size(); i++) {
      VobjDataHeader vdh = {0, -1, i};
      
      VobjList::const_iterator jend = vls[i].end();
      for (VobjList::const_iterator j = vls[i].begin(); j != jend; ++j) {
        const Vobj *vobj = *j;
//      const_cast<Vobj*>(vobj)->refreshData(); // --- already done above
        const VOBJSEG *vseg = vobj->getVseg();
        vdh.len = sizeof(W) + SIZE_VOBJSEG(vseg->dlen);

        writer.write(reinterpret_cast<B*>(&vdh), 
                     sizeof(VobjDataHeader));
        writer.write(reinterpret_cast<const B*>(vseg), 
                     SIZE_VOBJSEG(vseg->dlen));
      }
    }
    writer.flush();
  }

  // --- Ť VOBJSEG ǼѤΥ쥳ɤ

  for (;;) {
    W type = fnd_rec(fd, F_NBWD, (1 << VOBJSEGRECTYPE), VOBJSEGSUBTYPE, NULL);
    if (type < 0) {
      break;
    }
    del_rec(fd);
  }

  cls_fil(fd);

  changed = false;
}

void VobjSet::load(WID wid, const LINK* link)
{
  // ñ push back Ƥ OK

  W fd_link = -1;
  W fd_vseg = -1;
  W type, err;

  fd_link = opn_fil(const_cast<LINK*>(link), F_READ, NULL);
  fd_vseg = opn_fil(const_cast<LINK*>(link), F_READ, NULL);

  type = fnd_rec(fd_vseg, F_TOPEND, (1 << VOBJSEGRECTYPE), VOBJSEGSUBTYPE, NULL);
  if (type < 0) {
    cls_fil(fd_vseg);
    fd_vseg = -1;
  }

  BufferedReader *reader = 0;
  if (fd_vseg >= 0) {
    reader = new BufferedReader(fd_vseg);
  }

  for (W mode = F_TOPEND;; mode = F_NFWD) {
    type = fnd_rec(fd_link, mode, RM_LINK, 0, NULL);
    if (type < 0) {
      break;
    }

    VLINK target;
    VobjDataHeader vdh;
    std::vector<B> buf;

    // --- VLINK
    rea_rec(fd_link, 0, (B*)&target, sizeof(LINK), NULL, NULL);

    // --- VOBJSEG
    if (reader != 0) {
      err = reader->read(reinterpret_cast<B*>(&vdh), sizeof(VobjDataHeader)); 
      if (err != sizeof(VobjDataHeader)) {
        goto failure;
      }

      buf.resize(vdh.len - sizeof(W));

      err = reader->read(&(buf[0]), buf.size());
      if (err != buf.size()) {
        goto failure;
      }

      VID vid = oreg_vob(&target, &(buf[0]), wid, V_NODISP);
      Vobj* vobj = new Vobj(vid, target, reinterpret_cast<VOBJSEG*>(&(buf[0])));

      vls[vdh.page].push_back(vobj);
      continue;
    }
    else {
      goto failure;
    }

    continue; // --- ޤäϡfailure ʲϼ¹Ԥʤ
failure:
    if (reader != 0) {
      delete reader;
      reader = 0;
    }
    if (fd_vseg >= 0) {
      cls_fil(fd_vseg);
      fd_vseg = -1;
    }

    VOBJSEG default_vseg = 
      { {{0, 0, 0, 0}}, 0, 0, 0x10000000, 0x10000000, 
        0x10ffffff, 0x10ffffff, 0 };

    VID vid = oreg_vob(&target, &default_vseg, wid, V_NODISP);
    Vobj* vobj = new Vobj(vid, target, &default_vseg);
  
    vls[0].push_back(vobj);
  }

  if (fd_link >= 0) {
    cls_fil(fd_link);
  }
  if (reader != 0) {
    delete reader;
  }
  if (fd_vseg >= 0) {
    cls_fil(fd_vseg);
  }
}

