// Vobj Selection

#ifndef VOBJSEL_H_DEFINE
#define VOBJSEL_H_DEFINE

#include <basic.h>
#include <btron/vobj.h>
#include <btron/hmi.h>
#include <map>
#include "Vobj.h"

class VobjSelection {

  struct SelectedItem {
    Vobj* vobj;
    SEL_RGN region;
    
    SelectedItem(Vobj* vobj, W sts, double zoom)
      : vobj(vobj) {
      region.sts = sts;
      calcZoom(zoom);
    }

    void calcZoom(double zoom) {
      const RECT& org = vobj->getRect();
      region.rgn.r.c.left = (int)(org.c.left * zoom / 100.0);
      region.rgn.r.c.top = (int)(org.c.top * zoom / 100.0);
      region.rgn.r.c.right = (int)(org.c.right * zoom / 100.0);
      region.rgn.r.c.bottom = (int)(org.c.bottom * zoom / 100.0);
    }
  };

  typedef std::map<VID, SelectedItem> SelectionType;
  SelectionType selection;

  GID gid;
  double currentZoom;

protected:

  W getCurrentMode() const;

  void setAll(GID tgid, int mode);

  void dragZoom(GID dgid, W vcode, int dx, int dy);

  bool pushTmpTray(WEVENT *ev, VobjSet* vobjs, int page, bool remove, const PNT& firstp, double zoom);

  void robjCopyToTray(WEVENT *ev, const PNT& firstp, double zoom);

public:

  VobjSelection(GID gid, double zoom) : gid(gid), currentZoom(zoom) {}

  virtual ~VobjSelection() {}

  void append(Vobj* vobj);

  void remove(VID vid);

  void clear();

  Vobj* findVobj(const PNT& p);

  void move(WID wid, W vcode, int dx, int dy, double zoom);

  bool isSelected(VID vid) const;

  void drawFlicker(int mode) const;

  bool changeZoom(double zoom);

  void mouse_down(WID wid, WEVENT* ev, int page, double zoom, VobjSet* vobjs, Vobj* vobj);

  void deleteFromPage(VobjSet* vobjs, int page);

  void vobjCopyToPage(VobjSet* vobjs, int page);

  bool robjCopyToPage(VobjSet* vobjs, int page);

  void robjCopyByMenu(VobjSet* vobjs, const W* vlist, int page);

  UB* allocVidList() const;

  VID getMenuMode() const;

  bool updateVID(VID vid);

};

#endif // VOBJSEL_H_DEFINE
