//========================================================================
//
// XPDFApp.cc
//
// Copyright 2002-2003 Glyph & Cog, LLC
//
//========================================================================

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include "GString.h"
#include "GList.h"
#include "Error.h"
#include "XPDFViewer.h"
#include "XPDFApp.h"
#include "config.h"

#include <btron/vobj.h>
#include <btron/tf.h>
#include <btron/libapp.h>
#include <tstring.h>
#include "SplashBitmap.h"
#include "databox.h"

// these macro defns conflict with xpdf's Object class
#ifdef LESSTIF_VERSION
#undef XtDisplay
#undef XtScreen
#undef XtWindow
#undef XtParent
#undef XtIsRealized
#endif

#include <btron/proctask.h>
#include <keycode.h>
#include <errcode.h>
#include <btron/libapp.h>
#include "PDFDoc.h"
#include "fusen.h"

#include "toriaezu.h"
typedef char* String;

extern DEV_SPEC SCREEN;

extern XpdfFusen fsn;

#define WHEEL_SCROLL_OFFSET 10
#define CLK_SCROLL_OFFSET 5
#define DCLK_SCROLL_OFFSET 100

//------------------------------------------------------------------------

#define remoteCmdSize 512

//------------------------------------------------------------------------

#define nXOpts (sizeof(xOpts) / sizeof(XrmOptionDescRec))

#define nXResources (sizeof(xResources) / sizeof(XtResource))

//------------------------------------------------------------------------
// XPDFApp
//------------------------------------------------------------------------

SplashColor XPDFApp::paperRGB = {0xff, 0xff, 0xff};

XPDFApp::XPDFApp(int *argc, char *argv[])
{
  viewers = new GList();
  title = 0;

  bbar[3] = rbar[2] = 0; 
  setrect(zoombtn_r, CTL_ZOOMBTN_X, CTL_WINDOW_GAP, 
          CTL_ZOOMBTN_X + CTL_ZOOMBTN_WIDTH - 1, CTL_SWITCH_BOTTOM - 1);

  getscreen();
  initstdpnl();
}

XPDFApp::~XPDFApp() {
  delete title;
  deleteGList(viewers, XPDFViewer);
}

XPDFViewer *XPDFApp::openLink(LINK *alink, int page,
                              GString *ownerPassword, GString *userPassword) {
  TC fname[21];
  char s[41];
  XPDFViewer *viewer;

  link = *alink;

  fil_sts(alink, fname, NULL, NULL);
  tcstoeucs(s, fname);
  title = new GString(s);

  viewer = new XPDFViewer(this, alink, page, NULL,
			  ownerPassword, userPassword);
  viewers->append(viewer);
  return viewer;
}

void XPDFApp::close(XPDFViewer *viewer, GBool closeLast) {
  int i;

  if (viewers->getLength() == 1) {
    if (viewer != (XPDFViewer *)viewers->get(0)) {
      return;
    }
    if (closeLast) {
      quit();
    } else {
      viewer->clear();
    }
  } else {
    for (i = 0; i < viewers->getLength(); ++i) {
      if (((XPDFViewer *)viewers->get(i)) == viewer) {
	viewers->del(i);
	delete viewer;
	return;
      }
    }
  }
}

void XPDFApp::quit() {
  while (viewers->getLength() > 0) {
    XPDFViewer *viewer = (XPDFViewer*)viewers->get(0);

    if (viewer != 0 && viewer->core != 0) {
      fsn.zoomx100 = (int)(viewer->core->getZoom() * 100);
      fsn.page = viewer->core->getPageNum();
    }

    delete (XPDFViewer *)viewers->del(0);
  }

  cont = false;
}

void XPDFApp::quitQuery() 
{
  XPDFViewer *viewer = 0;
  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }

  VobjSet* vobjs = viewer->getVobjSet();

  if (vobjs->isChanged()) {
    W sel = panel(PNL_UPDEXIT);
    if (sel == 2) {
      gset_ptr(PS_BUSY, NULL, (COLOR)-1, (COLOR)-1);          
      vobjs->save(&link);
      gset_ptr(PS_SELECT, NULL, (COLOR)-1, (COLOR)-1);
    }
    if (sel != 0) {
      quit();
    }
  }
  else {
    quit();
  }
}


void XPDFApp::redisp(WEVENT *ev, XPDFViewer* view = 0) 
{
  XPDFViewer *viewer = view;

  if (viewer == 0 && viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }
  WID wid = viewer->getWID();

  // TODO: redisp ǤϤʤɥѷԤäʬǤ٤
  RECT r;
  wget_wrk(ev->s.wid, &r);
  setWorkArea(&r);

  do {
    if (wsta_dsp(wid, NULL, NULL) == 0)
      break;
    redraw(ev, viewer);
  } while (wend_dsp(wid) > 0);
}


void XPDFApp::redraw(WEVENT *ev, XPDFViewer* view = 0) 
{
  XPDFViewer *viewer = view;

  if (viewer == 0 && viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }
  WID wid = viewer->getWID();
  GID gid = wget_gid(wid);

  RECT r;
  wget_wrk(wid, &r);

  viewer->getVobjSel()->drawFlicker(0);

  bool exists_doc = 
    (viewer != 0 && viewer->core != 0 && viewer->core->getDoc() != 0);
  if (exists_doc) {
    if (r.c.right - r.c.left > bbar[2] - bbar[3]){
      RECT r2 = {{ bbar[2] - bbar[3], 0, r.c.right, r.c.bottom }};
      gfil_rec(gid, r2, viewer->getBackground(), 0, G_STORE);
    }
    if (r.c.bottom - r.c.top > rbar[3] - rbar[2]){
      RECT r2 = {{ 0, rbar[3] - rbar[2], r.c.right, r.c.bottom }};
      gfil_rec(gid, r2, viewer->getBackground(), 0, G_STORE);
    }
    viewer->updateScreen(gid);
    viewer->drawVobj();
  }
  else {
    wera_wnd(wid, NULL);
  }
}


void XPDFApp::redisp_sub(WEVENT *ev)
{
  XPDFViewer *viewer = 0;
  char spage[16], szoom[16];
  TC tspage[32], tszoom[32];
  static FSSPEC fs = {{TNULL}, FTC_DEFAULT, FT_FIXED, {8, 16}};
  SCRIPT scr;
  
  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }
  WID wid = viewer->getSubWID();
  GID gid = wget_gid(wid);

  if (viewer != 0 && viewer->core != 0 && viewer->core->getDoc() != 0) {
    sprintf(spage, "/%4d", viewer->core->getDoc()->getNumPages());
    eucstotcs(tspage, spage);
    sprintf(szoom, "%4d %%", (int)(viewer->core->getZoom()));
    eucstotcs(tszoom, szoom);
  }
  else {
    eucstotcs(tspage, "/    ");
    tszoom[0] = TNULL;
  }

  gset_fon(gid, &fs);
  gset_chc(gid, 0x10000000, 0x90ffffff);

  do {
    if (wsta_dsp(wid, NULL, NULL) == 0)
      break;
    wera_wnd(wid, NULL);
    scr = gdra_stp(gid, CTL_MAXPAGE_X, 16 + 8, 
                          tspage, 5, G_STORE);
    if (scr > 0) {
      gdra_chr(gid, (TSC_SYS | 0xfe00), G_STORE);
    }

    // zoom button
    gdra_lin(gid, zoombtn_r.p.lefttop, (PNT){zoombtn_r.c.left, zoombtn_r.c.bottom}, ((LN_SOLID << 8) | 1), WHITE0, G_STORE);
    gdra_lin(gid, zoombtn_r.p.lefttop, (PNT){zoombtn_r.c.right, zoombtn_r.c.top}, ((LN_SOLID << 8) | 1), WHITE0, G_STORE);
    gdra_lin(gid, zoombtn_r.p.rightbot, (PNT){zoombtn_r.c.left, zoombtn_r.c.bottom}, ((LN_SOLID << 8) | 1), BLACK100, G_STORE);
    gdra_lin(gid, zoombtn_r.p.rightbot, (PNT){zoombtn_r.c.right, zoombtn_r.c.top}, ((LN_SOLID << 8) | 1), BLACK100, G_STORE);
    gdra_lin(gid, (PNT){zoombtn_r.c.right-1, zoombtn_r.c.bottom-1}, (PNT){zoombtn_r.c.left+1, zoombtn_r.c.bottom-1}, ((LN_SOLID << 8) | 1), BLACK100, G_STORE);
    gdra_lin(gid, (PNT){zoombtn_r.c.right-1, zoombtn_r.c.bottom-1}, (PNT){zoombtn_r.c.right-1, zoombtn_r.c.top+1}, ((LN_SOLID << 8) | 1), BLACK100, G_STORE);
    scr = gdra_stp(gid, CTL_ZOOMBTN_X + 2, 16 + 8, 
                          tszoom, 6, G_STORE);
    if (scr > 0) {
      gdra_chr(gid, (TSC_SYS | 0xfe00), G_STORE);
    }

    // other controls
    cdsp_pwd(wid, NULL, P_RDISP);
  } while (wend_dsp(wid) > 0);
}

void XPDFApp::mouse_down(WEVENT *ev)
{
  XPDFViewer *viewer = 0;
  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }

  W clt;
  if (ev->s.cmd != W_WORK && ev->s.cmd != W_BBAR 
      && ev->s.cmd != W_RBAR && ev->s.cmd != W_LBAR) {
    clt = wchk_dck(ev->s.time);
  }

  switch (ev->s.cmd){
  case W_WORK:
    {
      Vobj* vobj = viewer->findVobj(ev->s.pos);
      if (vobj != 0) {
        mouse_down_vobj(ev, vobj);
      }
      else {
        viewer->getVobjSel()->clear();
      }
    }
    break;
    
  case W_LTHD:
  case W_LBHD:
  case W_RTHD:
  case W_RBHD:
    if (clt == W_DCLICK){
      if (viewer != 0) {
        wget_sts(ev->s.wid, &(viewer->rectsav), NULL);
        W err = wchg_wnd(ev->s.wid, NULL, W_MOVE);
        if (err == 0 || err == 2) {
          fsn.full = 0;
        }
        else if (err == 1) {
          fsn.full = 1;
        }
        if (err > 0) {
          redisp(ev);
        }
      }
    }
    else if (clt == W_PRESS || clt == W_QPRESS){
      wrsz_drg(ev, NULL, NULL); // update flag..
      fsn.full = 0;
      redisp(ev);
    }
    break;
    
  case W_PICT:
    if (clt == W_DCLICK){
      quitQuery();
    }
    else if (clt == W_CLICK){
      break; // no-operation
    }
  case W_TITL:
  case W_FRAM:
    if (clt == W_PRESS || clt == W_QPRESS){
      if (wmov_drg(ev, NULL) > 0){
        fsn.full = 0;
        redisp(ev);
      }
    }
    break;
    
  case W_LBAR:
    break;
  case W_RBAR:
  {
    if (rbar[1] == rbar[2] && rbar[0] == rbar[3]) {
      break;
    }
    RECT rect;
    W rbar_pid, act;
    wget_bar(ev->s.wid, &rbar_pid, NULL, NULL);
    for (bool cont = true; cont;) {
      act = cact_par(rbar_pid, ev);
      cget_val(rbar_pid, 2, rbar);
      switch (act){
          case 0x6000: // up(press)
            incr_rbar(ev->s.wid, rbar_pid, CLK_SCROLL_OFFSET, true);
            break;
          case 0x6001: // down(press)
            incr_rbar(ev->s.wid, rbar_pid, CLK_SCROLL_OFFSET, false);
            break;
          case 0x5004: // up(dck)
            incr_rbar(ev->s.wid, rbar_pid, DCLK_SCROLL_OFFSET, true);
            cont = false;
            break;
          case 0x5005: // down(dck)
            incr_rbar(ev->s.wid, rbar_pid, DCLK_SCROLL_OFFSET, false);
            cont = false;
            break;
          case 0x5008: // drag(up) - non P_DRAGBREAK
          case 0x5009: // drag(down) - non P_DRAGBREAK
            cont = false;
            break;
          case 0x6008: // drag(up) - P_DRAGBREAK
          case 0x6009: // drag(down) - P_DRAGBREAK
            rect.c.left = bbar[1];
            rect.c.top = rbar[0];
            setWorkArea(&rect);
            redraw(ev);
            break;
          default:
            cont = false;
      }
    }
    rect.c.left = bbar[1];
    rect.c.top = rbar[0];
    setWorkArea(&rect);
    redraw(ev);
    break;
  }
  case W_BBAR:
  {
    if (bbar[0] == bbar[3] && bbar[1] == bbar[2]){
      break;
    }
    RECT rect;
    W bbar_pid, act;
    wget_bar(ev->s.wid, NULL, &bbar_pid, NULL);
    for (bool cont = true; cont;) {
      act = cact_par(bbar_pid, ev);
      cget_val(bbar_pid, 2, bbar);
      switch (act){
          case 0x6002: // left(press)
            incr_bbar(ev->s.wid, bbar_pid, CLK_SCROLL_OFFSET, true);
            break;
          case 0x6003: // right(press)
            incr_bbar(ev->s.wid, bbar_pid, CLK_SCROLL_OFFSET, false);
            break;
          case 0x5006: // left(dck)
            incr_bbar(ev->s.wid, bbar_pid, DCLK_SCROLL_OFFSET, true);
            cont = false;
            break;
          case 0x5007: // right(dck)
            incr_bbar(ev->s.wid, bbar_pid, DCLK_SCROLL_OFFSET, false);
            cont = false;
            break;
          case 0x500a: // drag(left) - non P_DRAGBREAK
          case 0x500b: // drag(right) - non P_DRAGBREAK
            cont = false;
            break;
          case 0x600a: // drag(left) - P_DRAGBREAK
          case 0x600b: // drag(right) - P_DRAGBREAK
            rect.c.left = bbar[1];
            rect.c.top = rbar[0];
            setWorkArea(&rect);
            redraw(ev);
            break;
          default:
            cont = false;
      }
    }
    rect.c.left = bbar[1];
    rect.c.top = rbar[0];
    setWorkArea(&rect);
    redraw(ev);
    break;
  }

  case W_SWITCH:
    // ΰΰ
    break;
    
  case W_CLOSED:
    // ҥɥĤ
    break;
    
  }
}

void XPDFApp::mouse_down_vobj(WEVENT *ev, Vobj* vobj)
{
  XPDFViewer *viewer = 0;
  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }

  viewer->getVobjSel()->mouse_down(viewer->getWID(), ev, viewer->core->getPageNum(), 
                                   viewer->core->getZoom(), viewer->vobjs, vobj);
}

void XPDFApp::mouse_down_sub(WEVENT *ev)
{
  W pid, clt;
  if (ev->s.cmd != W_WORK) { 
    clt = wchk_dck(ev->s.time);
  }

  switch (ev->s.cmd){
  case W_WORK:
    if (cfnd_par(ev->s.wid, ev->s.pos, &pid) > 0) {
      XPDFViewer *viewer = 0;
      if (viewers->getLength() > 0) {
        viewer = (XPDFViewer*)viewers->get(0);
      }

      if (pid == viewer->sw_fwd_id) {
        // Υڡ
        if (cact_par(pid, ev) == 0x5001) {
          int p = viewer->core->getPageNum();
          viewer->core->gotoNextPage(1, gTrue);
          viewer->setPageDisp(p, viewer->core->getZoom());
        }
      }
      else if (pid == viewer->sw_bwd_id) {
        // Υڡ
        if (cact_par(pid, ev) == 0x5001) {
          int p = viewer->core->getPageNum();
          viewer->core->gotoPrevPage(1, gTrue, gFalse);
          viewer->setPageDisp(p, viewer->core->getZoom());
        }
      }
      else if (pid == viewer->sw_fst_id) {
        // ǽΥڡ
        if (cact_par(pid, ev) == 0x5001) {
          int p = viewer->core->getPageNum();
          viewer->core->gotoPrevPage(p - 1, gTrue, gFalse);
          viewer->setPageDisp(p, viewer->core->getZoom());
        }
      }
      else if (pid == viewer->sw_lst_id) {
        // ǸΥڡ
        if (cact_par(pid, ev) == 0x5001) {
          int maxp = viewer->core->getDoc()->getNumPages();
          int p = viewer->core->getPageNum();
          viewer->core->gotoNextPage(maxp - p, gTrue);
          viewer->setPageDisp(p, viewer->core->getZoom());
        }
      }
      else if (pid == viewer->nb_page_id) {
        // ڡֹ(ͥܥå)
        W sts = cact_par(pid, ev);
        if (sts == 0x5000 || sts == 0x5001) {
          int maxp = viewer->core->getDoc()->getNumPages();
          int p = viewer->core->getPageNum();
          W np;

          cget_val(pid, 1, &np);
          if (np > maxp) {
            np = maxp;
          }
          else if (np < 1) {
            np = 1;
          }
          cset_val(pid, 1, &np);

          int diff = np - p;
          if (diff > 0) {
            viewer->core->gotoNextPage(diff, gTrue);
            viewer->setPageDisp(p, viewer->core->getZoom());
          }
          else if (diff < 0) {
            viewer->core->gotoPrevPage(-diff, gTrue, gFalse);
            viewer->setPageDisp(p, viewer->core->getZoom());
          }
        }
      }
    }
    else if (inrect(zoombtn_r, ev->s.pos)) {
      XPDFViewer *viewer = 0;
      if (viewers->getLength() > 0) {
        viewer = (XPDFViewer*)viewers->get(0);
      }
      double prevZoom = viewer->core->getZoom();
      W mno = 
        viewer->getZoomMenu()->
        popupMenu(viewer->getSubWID(), prevZoom);
      if (mno > 0) {
        viewer->core->zoom = viewer->getZoomMenu()->getZoomVal(mno);
        viewer->setPageDisp(viewer->core->getPageNum(), prevZoom);
      }
    }
    break;

  case W_LTHD:
  case W_LBHD:
  case W_RTHD:
  case W_RBHD:
  case W_PICT:
  case W_TITL:
  case W_FRAM:
  case W_LBAR:
  case W_RBAR:
  case W_BBAR:
    if (clt == W_PRESS || clt == W_QPRESS){ // ɲ
      if (wmov_drg(ev, NULL) > 0){
        redisp_sub(ev);
      }
    }
    break;

  case W_SWITCH:
    // ΰΰ
    break;
    
  case W_CLOSED:
    // ҥɥĤ
    break;
    
  }
}

void XPDFApp::run() {
  WEVENT wev0;
  XPDFViewer *viewer = 0;

  if (viewers->getLength() > 0) {
    RECT r;
    viewer = (XPDFViewer*)viewers->get(0);
    wget_wrk(viewer->getWID(), &r);
    setWorkArea(&r);
    redraw(NULL, viewer);
    wreq_dsp(viewer->getSubWID());
  }

  // ٥ȥ롼: 㳰ȤΤǼ
  for (cont = true; cont; ) {
    wget_evt(&wev0, WAIT);
    
    switch (wev0.s.type){
        case EV_NULL:
        case EV_NOMSG:
          if (wev0.s.cmd == W_WORK) {
            if ((wev0.s.stat & (ES_LSHFT | ES_RSHFT)) != 0) {
              gset_ptr(PS_MODIFY, NULL, (COLOR)-1, (COLOR)-1);
            }
            else {
              Vobj* vobj = viewer->findSelectedVobj(wev0.s.pos);
              if (vobj != 0) {
                PNT zoomp;
                VID vid = vobj->getVID();
                zoomp.x = wev0.s.pos.x * 100.0 / viewer->core->getZoom();
                zoomp.y = wev0.s.pos.y * 100.0 / viewer->core->getZoom();
                W vcode = ofnd_vob(-vid, zoomp, &vid);
                if (vcode == V_LTHD || vcode == V_RTHD
                    || vcode == V_LBHD || vcode == V_RBHD) {
                  gset_ptr(PS_RSIZ, NULL, (COLOR)-1, (COLOR)-1);
                }
                else {
                  gset_ptr(PS_MOVE, NULL, (COLOR)-1, (COLOR)-1);
                }
              }
              else {
                gset_ptr(PS_SELECT, NULL, (COLOR)-1, (COLOR)-1);
              }
            }
          }
          viewer->getVobjSel()->drawFlicker(-1);
          break;
        case EV_REQUEST:
          switch (wev0.g.cmd) {
              case W_REDISP:
                if (wev0.s.wid == viewer->getWID()) {
                  redisp(&wev0);
                }
                else if (wev0.s.wid == viewer->getSubWID()) {
                  redisp_sub(&wev0);
                }
                break;
              case W_VOBJREQ:
                viewer->drawVobj();
                break;
              case W_PASTE:
                wrsp_evt(&wev0, paste(&wev0));
                break;
              case W_DELETE:
                wrsp_evt(&wev0, 0); // ACK
                quit();
                break;
              case W_FINISH:
                wrsp_evt(&wev0, 0); // ACK
                viewer->getVobjSet()->save(&link);
                quit();
                break;
          }
          break;
        case EV_RSWITCH:
          redisp(&wev0);
          redisp_sub(&wev0);
          /* FALLTHROUGH */
        case EV_SWITCH:
        case EV_BUTDWN:
          if (wev0.s.wid == viewer->getWID()) {
            mouse_down(&wev0);
          }
          else if (wev0.s.wid == viewer->getSubWID()) {
            mouse_down_sub(&wev0);
          }
          break;
        case EV_BUTUP:
          break;
        case EV_INACT:
          break;
        case EV_DEVICE:
          oprc_dev(&(wev0.e), NULL, 0);
          break;
        case EV_MSG:
          clr_msg(MM_ALL, MM_ALL);
          break;
        case EV_MENU:
        {
          W mno = viewer->getMenu()->popupMenu(&wev0, viewer->getVobjSel());
          if (mno > 0){
            selectMenu(&wev0, mno);
          }
          break;
        }
        case EV_KEYDWN:
          if (wev0.s.stat & ES_CMD) {
            W i = mfnd_key(viewer->getMenu()->getMenuID(), wev0.s.wid);
            if (i > 0){
              selectMenu(&wev0, i);
            }
            break;
          }
          /* FALLTHROUGH */
        case EV_AUTKEY:
          switch (wev0.e.data.key.code) {
              case KC_SS_D:
              case KC_SS_U:
                mwheel(&wev0);
                break;
              case KC_DEL:
                viewer->getVobjSel()->deleteFromPage(
                  viewer->getVobjSet(), viewer->core->getPageNum());
                redraw(&wev0, viewer);
                break;
              default:
                break;
          }
          break;
        case EV_KEYUP:
          break;
    }
  }
}

// retval = 0: ACK, 1: NACK
int XPDFApp::paste(WEVENT *ev)
{
  W size, nrec;
  PNT pos;
  TRAYREC *trayrec = 0;
  int retval = 1; // NACK
  // adsp_sel

  if (tget_dat(NULL, 0, &size, -1) < 0) {
    return -1;
  }

  trayrec = reinterpret_cast<TRAYREC*>(new UB[size]);
  if (trayrec == 0) {
    return -1;
  }

  if ((nrec = tget_dat(trayrec, size, &size, -1)) < 0) {
    return -1;
  }
  tset_dat(NULL, 0);

  XPDFViewer *viewer = 0;
  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }

  // Ψ100%ξΰ֤
  pos.x = ev->e.pos.x / viewer->core->getZoom() * 100.0 + .5;
  pos.y = ev->e.pos.y / viewer->core->getZoom() * 100.0 + .5;

  bool first = true;

  for (int i = 0; i < nrec; i++) {
    if (trayrec[i].id == TR_VOBJ) {
      Vobj* vobj = viewer->registerVobj
        (&(reinterpret_cast<TR_VOBJREC*>(trayrec[i].dt)->vlnk),
         &(reinterpret_cast<TR_VOBJREC*>(trayrec[i].dt)->vseg),
         pos, viewer->core->getPageNum());
      if (vobj != 0) {
        if (first) {
          // Ǥ򤵤ƤΤϡ٤˴
          viewer->getVobjSel()->clear();
          first = false;
        }
        viewer->getVobjSel()->append(vobj);
        retval = 0; // ACK
      }
    }
  }

  delete [] trayrec;

  //ev->e.pos.x = 0x8000;
  //ev->e.pos.y = 0x8000;

  return retval;
}

void XPDFApp::selectMenu(WEVENT *ev, W mno)
{
  XPDFViewer *viewer = 0;
  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }

  switch (viewer->getMenu()->getInternalID(mno)){
      case MN_FINISH: // λ
        quitQuery();
        break;
      case MN_NEWSAVE: // ¿Ȥ¸
        printf("MN_NEWSAVE\n");
        break;
      case MN_SAVE: // ¸
        {
          W sel = panel(PNL_UPDATE);
          if (sel == 1) {
            gset_ptr(PS_BUSY, NULL, (COLOR)-1, (COLOR)-1);          
            viewer->getVobjSet()->save(&link);
            gset_ptr(PS_SELECT, NULL, (COLOR)-1, (COLOR)-1);
          }
        }
        break;
      case MN_VOBJ: // 
        selectVobjMenu(ev, viewer, -(mno & 0xff));
        break;
      case MN_ROBJ: // ¿
        selectVobjMenu(ev, viewer, -((mno & 0xff) | 0x100));
        break;
      case MN_DISK: // ǥ
        selectVobjMenu(ev, viewer, -((mno & 0xff) | 0x200));
        break;
      case MN_REDISP: // ɽ
        wreq_dsp(viewer->getWID());
        wreq_dsp(viewer->getSubWID());
        break;
      case MN_WINDOW: // ɥ
        wexe_dmn(mno);
        break;
      case MN_GADGET: // ʪ
        oexe_apg(0, mno);
        break;
      case MN_EXEC: // ¹
        gset_ptr(PS_BUSY, NULL, (COLOR)-1, (COLOR)-1);
        if (oexe_apg(viewer->getVobjSel()->getMenuMode(), mno & 0xff) > 0) {
          gset_ptr(0x8001, NULL, (COLOR)-1, (COLOR)-1);
        }
        break;
  }
}

void XPDFApp::selectVobjMenu(WEVENT *ev, XPDFViewer *viewer, W item)
{
  typedef struct {
    W vid;
    RECT r;
  } OE_RLIST;

  VobjSet* vobjs = viewer->getVobjSet();
  VobjSelection* vsel = viewer->getVobjSel();

  UB* buf = vsel->allocVidList();
  W sel = oexe_vmn(0, item, buf);

  switch (sel) {
      case VM_OPEN:
      case VM_CLOSE:
      case VM_DISP:
        vobjs->setChanged();
        for (OE_RLIST* i = reinterpret_cast<OE_RLIST*>(buf); 
             *reinterpret_cast<W*>(i) != 0; i++) {
          vsel->updateVID(i->vid);
        }
        redraw(ev, viewer);
        break;
      case VM_NEW:
        vobjs->setChanged();
        vsel->robjCopyByMenu(vobjs, reinterpret_cast<W*>(buf), 
                             viewer->core->getPageNum());
        redraw(ev, viewer);
        break;
      case VM_RELN:
        /* FALLTHROUGH */
      case VM_NAME:
      case VM_DETACH:
      case VM_REFMT:
        // TODO: I_DLEDǤϤޤ˲ΰ׻vobjľƤ
        // ľ(ѹʤȤꤷƤ褵)
        vobjs->setChanged();
        redraw(ev, viewer);
        break;
      case VM_EXREQ:
        for (int i = 0; reinterpret_cast<W*>(buf)[i] != 0; i++) {
          oexe_apg(reinterpret_cast<W*>(buf)[i], 0);
        }
        break;
      case VM_PASTE:
        break;
      default:
        break;
  }

  delete [] buf;
}

void XPDFApp::incr_rbar(WID wid, W pid, int offset, bool up)
{
  if (up) {
    if (rbar[0] - offset < rbar[2]){
      offset = rbar[0] - rbar[2];
    }
    wscr_wnd(wid, NULL, 0, offset, W_SCRL | W_RDSET);
    rbar[0] -= offset;
    rbar[1] -= offset;
  }
  else {
    if (rbar[1] + offset > rbar[3]){
      offset = rbar[3] - rbar[1];
    }
    wscr_wnd(wid, NULL, 0, -offset, W_SCRL | W_RDSET);
    rbar[0] += offset;
    rbar[1] += offset;
  }
  cset_val(pid, 2, rbar);
}

void XPDFApp::incr_bbar(WID wid, W pid, int offset, bool left)
{
  if (left) {
    if (bbar[1] - offset < bbar[3]){
      offset = bbar[1] - bbar[3];
    }
    wscr_wnd(wid, NULL, offset, 0, W_SCRL | W_RDSET);
    bbar[1] -= offset;
    bbar[0] -= offset;
  }
  else {
    if (bbar[0] + offset > bbar[2]){
      offset = bbar[2] - bbar[0];
    }
    wscr_wnd(wid, NULL, -offset, 0, W_SCRL | W_RDSET);
    bbar[1] += offset;
    bbar[0] += offset;
  }
  cset_val(pid, 2, bbar);
}

void XPDFApp::normalizeScrollBars(WID wid)
{
  // normalize x-direction
  if (bbar[0] - bbar[1] >= bbar[2] - bbar[3]){
    bbar[1] = bbar[3];
    bbar[0] = bbar[2];
  }
  else {
    if (bbar[0] > bbar[2]){
      wscr_wnd(wid, NULL, bbar[0]-bbar[2], 0, W_MOVE | W_RDSET);
      bbar[1] -= bbar[0] - bbar[2];
      bbar[0] = bbar[2];
    }
    else if (bbar[3] > bbar[1]){
      wscr_wnd(wid, NULL, bbar[1]-bbar[3], 0, W_MOVE | W_RDSET);
      bbar[0] += bbar[3] - bbar[1];
      bbar[1] = bbar[3];
    }
  }
  // normalize y-direction
  if (rbar[1] - rbar[0] >= rbar[3] - rbar[2]){
    rbar[0] = rbar[2];
    rbar[1] = rbar[3];
  }
  else {
    if (rbar[1] > rbar[3]){
      wscr_wnd(wid, NULL, rbar[1]-rbar[3], 0, W_MOVE | W_RDSET);
      rbar[0] -= rbar[1] - rbar[3];
      rbar[1] = rbar[3];
    }
    else if (rbar[2] > rbar[0]){
      wscr_wnd(wid, NULL, rbar[2]-rbar[0], 0, W_MOVE | W_RDSET);
      rbar[1] += rbar[2] - rbar[0];
      rbar[0] = rbar[2];
    }
  }
}

void XPDFApp::setWorkArea(RECT* rect)
{
  XPDFViewer *viewer = 0;
  XPDFCore *core = 0;

  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }
  if (viewer == 0) {
    return;
  }
  WID wid = viewer->getWID();
  GID gid = wget_gid(wid);
  RECT wr;
  W bbar_pid, rbar_pid;

  wget_bar(wid, &rbar_pid, &bbar_pid, NULL);
  wget_wrk(wid, &wr);
  core = viewer->core;
  if (viewer->core != 0 && viewer->core->curTile != 0
      && viewer->core->curTile->bitmap != 0) {
    // Document Rect
    bbar[2] = bbar[3] + viewer->core->curTile->bitmap->getWidth();
    rbar[3] = rbar[2] + viewer->core->curTile->bitmap->getHeight();
    // ScrollBar Position
    bbar[1] = rect->c.left;
    rbar[0] = rect->c.top;
    bbar[0] = rect->c.left + (wr.c.right - wr.c.left);
    rbar[1] = rect->c.top + (wr.c.bottom - wr.c.top);
    normalizeScrollBars(wid);
    cset_val(bbar_pid, 4, bbar);
    cset_val(rbar_pid, 4, rbar);
    setrect(*rect, bbar[1], rbar[0], bbar[0], rbar[1]);
    wset_wrk(wid, rect);
    gset_vis(gid, *rect);
  }
}

void XPDFApp::mwheel(WEVENT *ev)
{
  XPDFViewer *viewer = 0;

  if (viewers->getLength() > 0) {
    viewer = (XPDFViewer*)viewers->get(0);
  }
  if (!viewer || (rbar[1] == rbar[2] && rbar[0] == rbar[3])) {
    return;
  }
  bool up = (ev->e.data.key.code == KC_SS_U);
  int n = ev->e.data.key.keytop - 0x8000;
  if (n < 0) n *= -1;
  if (n != 0) {
    W wid = viewer->getWID();
    W rbar_pid;
    wget_bar(wid, &rbar_pid, NULL, NULL);
    incr_rbar(wid, rbar_pid, WHEEL_SCROLL_OFFSET * n, up);
    redraw(ev);
  }
}

