//========================================================================
//
// XPDFApp.h
//
// Copyright 2002-2003 Glyph & Cog, LLC
//
//========================================================================

#ifndef XPDFAPP_H
#define XPDFAPP_H

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma interface
#endif

#include "gtypes.h"
#include "SplashTypes.h"

class GString;
class GList;
class XPDFViewer;

#include <btron/hmi.h>
#include <btron/vobj.h>
#include "toriaezu.h"
#include "Vobj.h"

//------------------------------------------------------------------------

#define xpdfAppName "̣Уģ"

//------------------------------------------------------------------------
// XPDFApp
//------------------------------------------------------------------------

class XPDFApp {
public:

  XPDFApp(int *argc, char *argv[]);
  ~XPDFApp();

  XPDFViewer *openLink(LINK *link, int page = 1,
		   GString *ownerPassword = NULL,
		   GString *userPassword = NULL);
  void close(XPDFViewer *viewer, GBool closeLast);
  void quit();
  void quitQuery();

  void run();

  GString *getTitle() { return title; }

  //----- resource/option values
  static GBool getInstallCmap() { return gFalse; }
  static int getRGBCubeSize() { return 5; }
  static GBool getReverseVideo() { return gFalse; }
  static SplashColorPtr getPaperRGB() { return paperRGB; }
  static Gulong getPaperPixel() { return 1; /* color index */ }
  static Gulong getMattePixel() { return 0; /* color index */ }

private:

  GList *viewers;		// [XPDFViewer]
  static SplashColor paperRGB;

  void redisp(WEVENT *ev, XPDFViewer *view = 0);
  void mouse_down(WEVENT *ev);
  void mouse_down_vobj(WEVENT *ev, Vobj* vobj);
  void mwheel(WEVENT *ev);
  int paste(WEVENT *ev);

  void redisp_sub(WEVENT *ev);
  void mouse_down_sub(WEVENT *ev);
  void selectMenu(WEVENT *ev, W mno);
  void selectVobjMenu(WEVENT *ev, XPDFViewer *viewer, W item);

  W rbar[4], bbar[4];
  RECT zoombtn_r;

  void incr_rbar(W wid, W pid, int offset, bool up);
  void incr_bbar(W wid, W pid, int offset, bool left);

  bool cont;
  GString *title;

  LINK link;

public:

  void redraw(WEVENT *ev, XPDFViewer *view = 0);

  void normalizeScrollBars(WID wid);
  void setWorkArea(RECT* rect);

  GList *getViewers() { return viewers; }

};

#endif
