//========================================================================
//
// XPDFViewer.h
//
// Copyright 2002-2003 Glyph & Cog, LLC
//
//========================================================================

#ifndef XPDFVIEWER_H
#define XPDFVIEWER_H

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma interface
#endif

#include "gtypes.h"
#include "XPDFCore.h"

#ifndef DISABLE_OUTLINE
#define DISABLE_OUTLINE
#endif

class GString;
class GList;
class UnicodeMap;
class LinkDest;
class XPDFApp;

#include <basic.h>
#include <btron/hmi.h>
#include <btron/dp.h>
#include <btron/vobj.h>
#include "toriaezu.h"
#include "MainMenu.h"
#include "ZoomMenu.h"
#include "Vobj.h"
#include "VobjSel.h"

//------------------------------------------------------------------------

// NB: this must match the defn of zoomMenuBtnInfo in XPDFViewer.cc
#define nZoomMenuItems 10

//------------------------------------------------------------------------
// XPDFViewer
//------------------------------------------------------------------------

class XPDFViewer {
public:

  XPDFViewer(XPDFApp *appA, LINK *link,
	     int pageA, GString *destName,
	     GString *ownerPassword, GString *userPassword);
  ~XPDFViewer();

  void open(LINK *link, int pageA, GString *destName);
  void clear();
  void reloadFile();

  WID getWindow() { return wid; }

  //----- load / display
  GBool loadFile(LINK *link, GString *ownerPassword = NULL,
		 GString *userPassword = NULL);
  void displayPage(int pageA, double zoomA, int rotateA,
                   GBool scrollToTop, GBool addToHist);
  void displayDest(LinkDest *dest, double zoomA, int rotateA,
		   GBool addToHist);
  void getPageAndDest(int pageA, GString *destName,
		      int *pageOut, LinkDest **destOut);

  //----- GUI code: main window
  void initWindow();
  void closeWindow();
//  int getZoomIdx();
//  void setZoomIdx(int idx);
  void setZoomVal(double z);

  void updateScreen(GID wgid);

  WID getWID() const { return wid; }
  WID getSubWID() const { return subwid; }
  PAT* getBackground() { return &background; }
  MainMenu* getMenu() { return &menu; }
  ZoomMenu* getZoomMenu() { return &zoommenu; }

  //----- Virtual Object
  Vobj* registerVobj(const VLINK* vlink, const VOBJSEG* vseg, 
                     const PNT& pos, int page);
  void drawVobj() const;
  VobjSelection* getVobjSel() { return vsel; }
  VobjSet* getVobjSet() { return vobjs; }
  Vobj* findVobj(const PNT& p);
  Vobj* findSelectedVobj(const PNT& p);

private:

  XPDFApp *app;

  WID wid;
  WID subwid;
  XPDFCore *core;
  TC *tc_title;
  PAT background;

  PAID sw_fst_id, sw_bwd_id, sw_fwd_id, sw_lst_id;
  PAID nb_page_id;
  RECT sw_fst_r, sw_bwd_r, sw_fwd_r, sw_lst_r;
  RECT nb_page_r;
  W nb_page_val;
  MainMenu menu;
  ZoomMenu zoommenu;

  WDSTAT rectsav; // -- ̥⡼ɤλˡɸФƤ

  VobjSet* vobjs;
  VobjSelection* vsel;

  // added for BTRON
  void setTitle(const char *s);
  void setPageDisp(int prevPage, double prevZoom);
  void setButtonAttr();

  friend class XPDFApp;
};

#endif
